/*
 * $Id: ResponseType.java 238 2005-08-15 22:26:43Z davidb $
 * 
 * Copyright (c) 2004 USC Project SRO. All rights reserved.
 * Copyright (c) 2005 VeriSign, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.iter;

import org.xbill.DNS.CNAMERecord;
import org.xbill.DNS.Flags;
import org.xbill.DNS.Name;
import org.xbill.DNS.Rcode;
import org.xbill.DNS.Section;
import org.xbill.DNS.Type;

import se.rfc.unbound.DelegationPoint;
import se.rfc.unbound.Request;
import se.rfc.unbound.SMessage;
import se.rfc.unbound.SRRset;

/**
 * This class represents a received DNS response. This encompasses the actual
 * message received as well as any ancillary data needed to processes or
 * interpret the response.
 * 
 * @author davidb
 * @version $Revision: 238 $
 */
public class ResponseType
{
  /** 'untyped' means that the type of this response hasn't been assigned. */
  public static final int UNTYPED   = 0;

  /** 'answer' means that the response terminates the resolution process. */
  public static final int ANSWER    = 1;
  /** 'delegation' means that the response is a delegation. */
  public static final int REFERRAL  = 2;
  /**
   * 'cname' means that the response is a cname without the final answer, and
   * thus must be restarted.
   */
  public static final int CNAME     = 3;
  /**
   * 'throwaway' means that this particular response should be discarded and
   * the next nameserver should be contacted
   */
  public static final int THROWAWAY = 4;

  /**
   * 'lame' means that this particular response indicates that the nameserver
   * knew nothing about the question.
   */
  public static final int LAME      = 5;


  /**
   * Classifies a response message from cache based on the current request.
   * Note that this routine assumes that THROWAWAY or LAME responses will not
   * occur. Also, it will not detect REFERRAL type messages, since those are
   * (currently) automatically classified based on how they came from the
   * cache (findDelegation() instead of lookup()).
   * 
   * @param m The response message to check.
   * @param request The current request that generated the response.
   * @return the response type (CNAME or ANSWER).
   */
  public static int typeCacheResponse(SMessage m, Request request)
  {
    // If the message is NXDOMAIN, then it is an ANSWER.
    if (m.getRcode() == Rcode.NXDOMAIN) return ANSWER;

    Name qname = request.getQName();
    int qtype = request.getQType();
    int qclass = request.getQClass();

    // First we look at the answer section. This can tell us if this is a
    // CNAME or postive ANSWER.

    SRRset[] answer_rrsets = m.getSectionRRsets(Section.ANSWER);

    if (answer_rrsets.length > 0)
    {
      // Now look at the answer section first. 3 states: our answer is there
      // directly, our answer is there after a cname, or there is just a
      // cname.
      Name mname = qname;

      for (int i = 0; i < answer_rrsets.length; i++)
      {
        SRRset s = answer_rrsets[i];

        // If we have encountered an answer (before or after a CNAME), then we
        // are done! Note that if qtype == CNAME then this will be noted as an
        // ANSWER before it gets treated as a CNAME, as it should.
        if (s.getType() == qtype && s.getDClass() == qclass
            && s.getName().equals(mname))
        {
          return ANSWER;
        }

        // If we have encountered a CNAME, make sure that it is relevant.
        if (s.getType() == Type.CNAME && s.getName().equals(mname))
        {
          mname = ((CNAMERecord) s.first()).getTarget();
        }
      }

      // if we encountered a CNAME (or a bunch of CNAMEs), and still got to
      // here, then it is a CNAME response. (i.e., the CNAME chain didn't
      // terminate in an answer rrset.)
      if (!mname.equals(qname)) return CNAME;
    }

    // At this point, since we don't need to detect REFERRAL or LAME messages,
    // it can only be an ANSWER.
    return ANSWER;
  }

  /**
   * Classifies a response message (from the wire) based on the current
   * request.
   * 
   * NOTE: currently this routine uses the AA bit in the response to help
   * distinguish between some non-standard referrals and answers. It also
   * relies somewhat on the originating zone to be accurate (for lameness
   * detection, mostly).
   * 
   * @param m The response message.
   * @param qname The original query name.
   * @param qtype The original query type.
   * @param qclass The original query class.
   * @return the response type.
   */
  public static int typeWireResponse(SMessage m, Request request, 
      DelegationPoint currentDelegationPoint)
  {
    if (m == null) return THROWAWAY;

    // If the message is NXDOMAIN, then it answers the question.
    if (m.getRcode() == Rcode.NXDOMAIN) return ANSWER;

    // Other response codes mean (so far) to throw the response away as
    // meaningless and move on to the next nameserver.
    if (m.getRcode() != Rcode.NOERROR) return THROWAWAY;

    // If the TC bit was set, then we generally have no choice but to retry
    // over TCP.
    if (m.getHeader().getFlag(Flags.TC)) return THROWAWAY;

    if (request == null) return 0;

    Name qname = request.getQName();
    int qtype = request.getQType();
    int qclass = request.getQClass();

    Name origZone = Name.root; // the default.
    if (currentDelegationPoint != null)
    {
      origZone = currentDelegationPoint.getName();
    }

    // First we look at the answer section. This can tell us if this is a
    // CNAME or ANSWER or (provisional) ANSWER.

    SRRset[] answer_rrsets = m.getSectionRRsets(Section.ANSWER);

    if (answer_rrsets.length > 0)
    {
      // Now look at the answer section first. 3 states: our answer is there
      // directly, our answer is there after a cname, or there is just a
      // cname.
      Name mname = qname;

      for (int i = 0; i < answer_rrsets.length; i++)
      {
        SRRset s = answer_rrsets[i];

        // If we have encountered an answer (before or after a CNAME), then we
        // are done! Note that if qtype == CNAME then this will be noted as an
        // ANSWER before it gets treated as a CNAME, as it should.
        if (s.getType() == qtype && s.getDClass() == qclass
            && s.getName().equals(mname))
        {
          if (m.getHeader().getFlag(Flags.AA)) return ANSWER;
          // If the AA bit isn't on, and we've seen the answer, we only
          // provisionally say 'ANSWER' -- it very well could be a REFERRAL.
          break;
        }

        // If we have encountered a CNAME, make sure that it is relevant.
        if (s.getType() == Type.CNAME && s.getName().equals(mname))
        {
          mname = ((CNAMERecord) s.first()).getTarget();
        }
      }

      // if we encountered a CNAME (or a bunch of CNAMEs), and still got to
      // here, then it is a CNAME response. (This is regardless of the AA bit
      // at this point)
      if (!mname.equals(qname)) return CNAME;
    }

    // Looking at the authority section, we just look and see if there is a
    // delegation NS set, turning it into a delegation. Otherwise, we will
    // have to conclude ANSWER (either it is NOERROR/NODATA, or an
    // non-authoritative answer)
    SRRset[] auth_rrsets = m.getSectionRRsets(Section.AUTHORITY);

    for (int i = 0; i < auth_rrsets.length; i++)
    {
      SRRset s = auth_rrsets[i];

      // The normal way of detecting NOERROR/NODATA.
      if (s.getType() == Type.SOA && qname.subdomain(s.getName()))
      {
        return ANSWER;
      }

      // Detect REFERRAL/LAME/ANSWER based on the relationship of the NS set
      // to the originating zone name.
      if (s.getType() == Type.NS)
      {
        // If we are getting an NS set for the zone we thought we were
        // contacting, then it is an answer.
        // FIXME: is this correct?
        if (s.getName().equals(origZone)) return ANSWER;
        // If we are getting a referral upwards (or to the same zone), then the server is 'lame'.
        if (origZone.subdomain(s.getName())) return LAME;

        // If the NS set is below the delegation point we are on, and it is
        // non-authoritative, then it is a referral, otherwise it is an
        // answer.
        if (s.getName().subdomain(origZone))
        {
          // NOTE: I no longer remember in what case we would like this to be an answer.
          // NODATA should have a SOA or nothing, not an NS rrset.
          // True, referrals should not have the AA bit set, but...
          
          // if (m.getHeader().getFlag(Flags.AA)) return ANSWER;
          return REFERRAL;
        }
        // Otherwise, the NS set is irrelevant.
      }
    }

    // If we've gotten this far, this is NOERROR/NODATA (which could be an
    // entirely empty message)
    return ANSWER;
  }

  public static String string(int responsetype)
  {
    switch (responsetype)
    {
      case ANSWER :
        return "ANSWER";
      case REFERRAL :
        return "REFERRAL";
      case CNAME :
        return "CNAME";
      case THROWAWAY :
        return "THROWAWAY";
      case LAME :
        return "LAME";
      default :
        return "UKNOWN response type";
    }
  }
}