/*
 * $Id: ServerUtils.java 238 2005-08-15 22:26:43Z davidb $
 *
 * Copyright (c) 2005 VeriSign, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package se.rfc.unbound.server;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectableChannel;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;

import se.rfc.unbound.*;

public class ServerUtils
{
  // the version.bind (version.server, whatever) string.
  private String           mVersionString     = "unbound 0.2-alpha";
  // the qname to actually match.
  private Name             mVersionQName;
  // a cached version of the version response in message format.
  private Message          mVersionMessage;

  // reusable udp read buffer
  private ByteBuffer       mUDPBuffer;
  // reusable tcp read buffer for the length bytes.
  private ByteBuffer       mTCPLengthBuffer;

  private Logger           log                = Logger.getLogger(this
                                                  .getClass());

  // this is what *we* advertise as a payload size.
  // FIXME: should be configurable.
  private static final int EDNS0_PAYLOAD_SIZE = 1280;
  // The maximum size of a UDP packet that we will accept.
  private static final int MAX_UDP_PAYLOAD    = 4096;

  public ServerUtils(Properties config)
  {
    mVersionString = config.getProperty("server.version", mVersionString);
    mVersionQName = Name.fromConstantString("version.bind.");

    mUDPBuffer = ByteBuffer.allocateDirect(MAX_UDP_PAYLOAD);
    mTCPLengthBuffer = ByteBuffer.allocateDirect(2);
  }

  /**
   * Format a generic error message with the given result code. Do not
   * generate an error message for requests that are marked as responses.
   * 
   * @param header the header to use.
   * @param rcode the rcode to use in the header.
   * 
   * @return an error response message matching the original header.
   */
  public Message errorMessage(Header header, int rcode)
  {
    if (header.getFlag(Flags.QR)) return null;

    Message m = new Message(header.getID());
    Header h = m.getHeader();
    h.setFlag(Flags.QR, true);
    // FIXME: we probably want to only set RA when recursion is actually
    // available.
    h.setFlag(Flags.RA, true);
    // FIXME: what other flags should be copied?

    h.setRcode(rcode);

    return m;
  }

  /**
   * Format a generic error message with the given result code
   * 
   * @param query the error producing query.
   * @param rcode the result code to use.
   * 
   * @return An error message matching the orignal query message.
   */
  public Message errorMessage(Message query, int rcode)
  {
    Message m = errorMessage(query.getHeader(), rcode);
    m.addRecord(query.getQuestion(), Section.QUESTION);

    return m;
  }

  /**
   * Format a generic error message with the given result code
   * 
   * @param in the wire-format query.
   * @param rcode the result code to use.
   * 
   * @return An error message matching the encoded DNS message.
   */
  public Message errorMessage(byte[] in, int rcode)
  {
    try
    {
      Header header = new Header(in);
      return errorMessage(header, rcode);
    }
    catch (IOException e)
    {
      log.info("Could not parse request message header", e);
      return null;
    }
  }

  public Message versionResponse(Message queryMessage)
  {
    Record question = queryMessage.getQuestion();
    if (question.getDClass() != DClass.CH || question.getType() != Type.TXT
        || !question.getName().equals(mVersionQName))
    {
      return null;
    }

    if (mVersionMessage == null)
    {
      mVersionMessage = new Message(0);
      TXTRecord version_txt = new TXTRecord(mVersionQName, DClass.CH, 0,
          mVersionString);
      NSRecord version_ns = new NSRecord(mVersionQName, DClass.CH, 0,
          mVersionQName);

      mVersionMessage.addRecord(version_txt, Section.ANSWER);
      mVersionMessage.addRecord(version_ns, Section.AUTHORITY);

      Header h = mVersionMessage.getHeader();
      h.setFlag(Flags.QR);
      h.setFlag(Flags.AA);
    }

    Message m = (Message) mVersionMessage.clone();
    Header h = m.getHeader();
    h.setID(queryMessage.getHeader().getID());
    h.setFlag(Flags.RD, queryMessage.getHeader().getFlag(Flags.RD));

    return m;
  }

  /**
   * @return the maximum payload size specified by EDNS0, or the default sizes
   *         for TCP or UDP
   */
  public int getMaximumPayloadSize(Message queryMessage, boolean tcp)
  {
    if (tcp)
    {
      return 65535;
    }

    OPTRecord opt = queryMessage.getOPT();
    if (opt != null)
    {
      return opt.getPayloadSize();
    }

    return 512;
  }

  /**
   * Given a response message and a (original) request, convert the response
   * message, forcing it to match the orignal request.
   * 
   * @param respMessage The response SMessage.
   * @param req The orignal request.
   * @return A possibly modifed Message, ready for wire formatting.
   */
  public Message formatResponseMessage(SMessage respMessage, Request req)
  {
    // Don't care about any OPT record in the response -- it probably came
    // from some auth server and is irrelevant.
    respMessage.setOPT(null);

    Message m = respMessage.getMessage();

    // Make the header match the request.
    Header h = m.getHeader();
    h.setID(req.getID());

    // copy certain header bits from the request to the response.
    h.setFlag(Flags.RD, req.getFlag(Flags.RD));
    h.setFlag(Flags.CD, req.getFlag(Flags.CD));

    // now handle the OPT rr.
    OPTRecord orig_opt = req.getOPT();
    if (orig_opt != null)
    {
      OPTRecord opt = new OPTRecord(EDNS0_PAYLOAD_SIZE, orig_opt
          .getExtendedRcode(), 0, orig_opt.getFlags(), orig_opt.getOptions());
      m.addRecord(opt, Section.ADDITIONAL);
    }

    return m;
  }

  /**
   * Validate a request message as a valid request.
   * 
   * @param queryMessage The request message.
   * @return null if the request was valid, otherwise return an error message.
   */
  public Message validateRequestMessage(Message queryMessage)
  {

    Record question = queryMessage.getQuestion();

    // Lack of a question should result in a format error.
    if (question == null)
    {
      return errorMessage(queryMessage, Rcode.FORMERR);
    }

    // For now, we only accept QUERY opcodes. In the future we may accept
    // others.
    if (queryMessage.getHeader().getOpcode() != Opcode.QUERY)
    {
      return errorMessage(queryMessage, Rcode.NOTIMP);
    }

    int qtype = queryMessage.getQuestion().getType();
    if (qtype == Type.OPT || qtype == Type.TKEY || qtype == Type.TSIG)
    {
      return errorMessage(queryMessage, Rcode.FORMERR);
    }

    if (qtype == Type.AXFR || qtype == Type.IXFR)
    {
      return errorMessage(queryMessage, Rcode.NOTIMP);
    }

    // Handle the version.bind query here.
    Message v = versionResponse(queryMessage);
    if (v != null) return v;

    return null;
  }

  /**
   * Convert a channel into a hopefully identifiable string. NOTE:
   * DatagramChannels can be difficult to identify property if not bound or
   * connect, which is typical.
   * 
   * @param ch The channel
   * @return A short, descriptive string.
   */
  public static String channelToString(SelectableChannel ch)
  {
    if (ch instanceof DatagramChannel)
    {
      DatagramChannel dch = (DatagramChannel) ch;
      DatagramSocket sock = dch.socket();
      if (sock.isBound())
      {
        try
        {
          SocketAddress local_addr = sock.getLocalSocketAddress();
          return "UDP(" + local_addr + ")";
        }
        catch (IllegalArgumentException e)
        {
          return "UDP()";
        }
      }
      return "UDP(? - " + dch.hashCode() + ")";
    }
    else if (ch instanceof SocketChannel)
    {
      SocketChannel sch = (SocketChannel) ch;
      SocketAddress local_addr = sch.socket().getLocalSocketAddress();
      if (local_addr != null)
      {
        return "TCP(" + local_addr + ")";
      }
      int port = sch.socket().getLocalPort();
      return "TCP(" + port + ")";
    }
    else if (ch instanceof ServerSocketChannel)
    {
      ServerSocketChannel sch = (ServerSocketChannel) ch;
      SocketAddress local_addr = sch.socket().getLocalSocketAddress();
      return "TCP*(" + local_addr + ")";
    }

    return "UNK()";
  }

  public static class ReadResult
  {
    public InetSocketAddress sourceAddress;
    public byte[]            in;
    public boolean           closeChannel;
  }

  public ReadResult readUDP(DatagramChannel dch) throws IOException
  {
    mUDPBuffer.clear();

    ReadResult rr = new ReadResult();
    rr.sourceAddress = (InetSocketAddress) dch.receive(mUDPBuffer);
    // if there was no packet after all, just return now.
    if (rr.sourceAddress == null)
    {
      // log.debug("Apparently, no packet was ready on channel --
      // returning");
      return null;
    }

    mUDPBuffer.flip(); // flip the buffer for reading
    rr.in = new byte[mUDPBuffer.limit()];
    mUDPBuffer.get(rr.in);
    return rr;
  }

  public ReadResult readTCP(ChannelState state, SocketChannel sch)
      throws IOException
  {
    ReadResult rr = new ReadResult();

    rr.sourceAddress = (InetSocketAddress) sch.socket()
        .getRemoteSocketAddress();

    ByteBuffer readBuffer = state.getReadBuffer();
    int n = 0;

    // if we didn't have a read buffer set, then this is our first time
    // through.
    // first we read in the length...
    if (readBuffer == null)
    {
      mTCPLengthBuffer.clear();

      // NOTE: we loop on the read() call, but we are only reading 2 bytes.
      while ((n = sch.read(mTCPLengthBuffer)) > 0)
      {
        // Nothing to do here -- it is all handled in the condition
        // expression.
      }

      if (n < 0)
      {
        log.debug("TCP channel closed, deregistering");
        rr.closeChannel = true;
        return rr;
      }

      mTCPLengthBuffer.flip();
      int len = mTCPLengthBuffer.getShort();
      readBuffer = ByteBuffer.allocateDirect(len);
      state.setReadBuffer(readBuffer);
    }

    while ((n = sch.read(readBuffer)) > 0)
    {
      // Nothing to do here -- it is all done in the condition expression.
    }
    log.debug("handling TCP read: read " + readBuffer.position() + " bytes");

    // if n ended < 0, then the socket actually closed on us, so we are
    // done.
    if (n < 0)
    {
      log.debug("TCP channel closed, deregistering");
      rr.closeChannel = true;
      return rr;
    }

    // Have we read the entire message?
    if (readBuffer.hasRemaining())
    {
      log.debug("partial read; returning");
      return null;
    }

    // Copy the data into our byte array.
    readBuffer.flip();
    rr.in = new byte[readBuffer.capacity()];
    readBuffer.get(rr.in);
    return rr;
  }

  public boolean writeUDP(WriteEvent w, DatagramChannel dch,
      ChannelState state) throws IOException
  {
    ByteBuffer writeBuffer = w.getData();
    int n = dch.send(writeBuffer, w.getDestination());

    if (n == 0)
    {
      log.debug("UDP write was incomplete.");
      state.unfinishedWriteEvent(w);
      return false;
    }
    return true;
  }

  public boolean writeTCP(WriteEvent w, SocketChannel sch,
      ChannelState state) throws IOException
  {
    ByteBuffer writeBuffer = w.getData();
    int n = 0;
    log.debug("Writing to TCP channel: " + sch);
    mTCPLengthBuffer.clear();
    mTCPLengthBuffer.putShort(0, (short) writeBuffer.capacity());
    n = sch.write(mTCPLengthBuffer);

    while ((n = sch.write(writeBuffer)) > 0)
    {
      log.debug("wrote " + n + " bytes of " + writeBuffer.capacity()
          + " to: " + w.getDestination());
    }

    if (writeBuffer.hasRemaining())
    {
      // our write didn't actually complete.
      log.debug("TCP write was incomplete.");
      state.unfinishedWriteEvent(w);
      return false;
    }
    return true;
  }
  
  /**
   * Load a particular ServerSelection implementation.
   * 
   * @param className The name of the DNSCache implementation.
   * @return an uninitialized instance of the DNSCache.
   * 
   * @throws Exception If something goes wrong with loading the class.
   */
  public static ServerSelection loadServerSelection(String className)
      throws Exception
  {
    if (className == null || className.equals("")) return null;

    Class c = null;

    c = Class.forName(className);

    Object o = null;

    o = c.newInstance();

    return (ServerSelection) o;
  }

}
